import numpy as np
from isbtchot import model
from isbtchot import view as v
from isbtchot.schemas.args import TypeTime


view = v.ViewPlotText()


def btc_terminal(time_grouping: TypeTime, periods_back: int):
    df = model.btc_hist(time_grouping, periods_back)
    view.candle_stick_terminal(df, "BTC/USD")
    view.plt.yscale("log")


def btc_pi_terminal(time_grouping: TypeTime, periods_back: int):
    df = model.btc_pi(time_grouping, periods_back)
    view.chart_terminal(df.time, df.pi, "Hotness Index")
    view.add_horizontal_line(1, "red")
    view.add_horizontal_line((0.95 - 0.35) / (1 - 0.35), "red")
    view.add_horizontal_line((0.80 - 0.35) / (1 - 0.35), "orange")
    view.add_horizontal_line((0.66 - 0.35) / (1 - 0.35), "green")
    view.add_horizontal_line(0, "blue")
    for _, row in df.iterrows():
        if row.pi_sell:
            view.add_vertical_line(row.time, "red")
        if row.pi_buy:
            view.add_vertical_line(row.time, "green")
    view.plt.ylim(df.pi.min(), max(1, df.pi.max()))


def btc_power_law_terminal(time_grouping: TypeTime, periods_back: int):
    df = model.btc_power_law(time_grouping, periods_back)
    delta = (np.log(df["delta"].dropna()) - np.log(0.45)) / 2
    days = df["days"][: len(delta)]

    # Plot data
    view.chart_terminal(days, delta, "Power Law Delta")
    view.add_horizontal_line(1, "red")
    view.add_horizontal_line((0 - np.log(0.45)) / 2, "orange")
    view.add_horizontal_line(0, "green")


# Dashboards
##################################################################


def dashboard_power_law(time_grouping: TypeTime, periods_back: int):
    view.dimensions(0, -4)
    view.subplots(2, 1)
    view.select_subplot(1, 1)
    btc_terminal(time_grouping, periods_back)
    view.select_subplot(2, 1)
    btc_power_law_terminal(time_grouping, periods_back)
    view.show()


def dashboard_isbtchot(time_grouping: TypeTime, periods_back: int):
    view.dimensions(0, -4)
    view.subplots(2, 1)
    view.select_subplot(1, 1)
    btc_terminal(time_grouping, periods_back)
    view.select_subplot(2, 1)
    btc_pi_terminal(time_grouping, periods_back)
    view.show()



import numpy as np
import requests
import datetime
import pandas as pd
import isbtchot
from isbtchot.schemas.args import TypeTime
from sklearn.linear_model import LinearRegression


CACHE_BTC_PATH = isbtchot.root_path / "cache" / "btc.csv"
BTC_API = (
    "https://min-api.cryptocompare.com/data/v2/histoday?fsym=BTC&tsym=USD&allData=true"
)


def btc_historical_daily() -> pd.DataFrame:
    # Ensure the directory exists
    CACHE_BTC_PATH.parent.mkdir(parents=True, exist_ok=True)

    # Check if the file exists and was modified today
    if (
        CACHE_BTC_PATH.is_file()
        and datetime.datetime.fromtimestamp(CACHE_BTC_PATH.stat().st_mtime).date()
        == datetime.datetime.today().date()
    ):
        df = pd.read_csv(CACHE_BTC_PATH)

    # Fetch new data
    else:
        try:
            data = requests.get(BTC_API).json()["Data"]["Data"]

        except Exception as _:
            # Try without ssl enabled
            data = requests.get(BTC_API, verify=False).json()["Data"]["Data"]

        df = pd.DataFrame(data)
        df.to_csv(CACHE_BTC_PATH, index=False)

    df["time"] = pd.to_datetime(df["time"], unit="s")
    df.set_index("time", inplace=True)
    return df


def btc_pi(time_grouping: TypeTime, periods_back: int | None = None):
    df = btc_historical_daily()[["close"]].rename({"close": "price"}, axis=1)

    df["sma111"] = df["price"].rolling(window=111).mean()
    df["sma350x2"] = df["price"].rolling(window=350).mean() * 2
    df["pi"] = df["sma111"] / df["sma350x2"]

    df = df.drop(["sma111", "sma350x2"], axis=1)
    df = df.dropna()

    # Normalize pi
    df["pi"] = (df["pi"] - 0.35) / (1 - 0.35)

    # Sell Indicator
    mask_pi_sell = (df["pi"].shift(-1) >= 1) & (df["pi"] < 1)
    df["pi_sell"] = mask_pi_sell

    # Buy indicator
    mask_pi_buy = (df["pi"].shift(-1) <= 0) & (df["pi"] > 0)
    df["pi_buy"] = mask_pi_buy

    df = df.resample(time_grouping.value).agg(
        {
            "price": "last",
            "pi": "max",
            "pi_sell": "max",
            "pi_buy": "max",
        }
    )

    df = df.reset_index()
    df.time = df.time.dt.strftime("%d/%m/%Y")

    if periods_back:
        periods_back = min(periods_back, len(df))
        df = df.iloc[-periods_back:]

    return df


def btc_hist(time_grouping: TypeTime | None = None, periods_back: int | None = None):
    df = btc_historical_daily()
    df = df.rename(
        {"open": "Open", "close": "Close", "high": "High", "low": "Low"}, axis=1
    )
    if time_grouping:
        df = df.resample(time_grouping.value).agg(
            {
                "Open": "first",
                "High": "max",
                "Low": "min",
                "Close": "last",
            }
        )
    if periods_back:
        df = df.iloc[-periods_back:]
    return df


def btc_power_law(
    time_grouping: TypeTime | None = None,
    periods_back: int | None = None,
    years_to_predict: int = 20,
) -> pd.DataFrame:
    df = btc_hist(time_grouping=time_grouping)[["Close"]].rename(
        {"Close": "price"}, axis=1
    )

    # Adjust for genesis block
    genesis_delta = (df.index[0] - pd.Timestamp(year=2009, month=1, day=3)).days
    if time_grouping is TypeTime.WEEK:
        genesis_delta /= 7
    elif time_grouping is TypeTime.MONTH:
        genesis_delta /= 30

    df["days"] = np.arange(len(df)) + genesis_delta
    # Applying log transformation
    X_log = np.log(df["days"].values.reshape(-1, 1))
    y_log = np.log(df["price"].values)

    # Fit the linear regression model on the log-transformed data
    power_law_model = LinearRegression().fit(X_log, y_log)

    # The coefficient 'b' is the slope of the line in the log-log space
    b = power_law_model.coef_[0]
    # The coefficient 'a' is obtained by taking the exponential of the intercept in the log-log space
    a = np.exp(power_law_model.intercept_)

    # Extend df
    periods = 365 * years_to_predict
    df = pd.concat(
        [
            df,
            pd.DataFrame(
                {
                    "days": np.arange(periods) + df.days.iloc[-1],
                },
                index=pd.date_range(df.index[-1], periods=periods, freq="D"),
            ),
        ]
    )

    # Using the power law model to predict prices across the original range of days
    df["power_law"] = a * (df.days**b)
    df["power_law_bottom"] = df.power_law * 0.45
    df["power_law_top"] = df.power_law * 3
    df["delta"] = df.price / df.power_law

    if periods_back:
        df = df.dropna().iloc[-periods_back:]
    return df



from pathlib import Path
import warnings

# Ignore specific warning categories, in this case, FutureWarning
warnings.filterwarnings("ignore", category=FutureWarning)

root_path: Path = Path(__file__).parent


from isbtchot import controller
from isbtchot.schemas.args import TypeTime
import argparse

parser = argparse.ArgumentParser(description="BTC Hotness Index", prog="isbtchot")

parser.add_argument(
    "-p",
    "--periods_back",
    type=int,
    default=85,
    help="Periods back to be processed. Defaults to 85 if not provided.",
)

parser.add_argument(
    "-t",
    "--time",
    type=TypeTime,
    default=TypeTime.WEEK,
    help="Candle stick time to use: ME or W. Defaults to W (Weekly)",
)

parser.add_argument(
    "-d",
    "--dashboard",
    type=str,
    default="isbtchot",
    help="Dashboard to display: 'isbtchot' or 'power_law'. Defaults to 'isbtchot'.",
)


def main():

    # Parse args
    args = parser.parse_args()
    time_grouping: TypeTime = args.time
    periods_back: int = args.periods_back
    dashboard: str = args.dashboard

    if dashboard == "isbtchot":
        controller.dashboard_isbtchot(
            periods_back=periods_back, time_grouping=time_grouping
        )
    elif dashboard == "power_law":
        controller.dashboard_power_law(
            periods_back=periods_back, time_grouping=time_grouping
        )
    else:
        raise ValueError(f"Unknown dashboard: '{dashboard}'. Expected 'isbtchot' or 'power_law'.")

if __name__ == "__main__":
    main()
